/* Pan - Article cache
 *
 * Copyright (C) 1999  Matt Eagleson (e.Messiah@superpimp.org)
 *
 * Authors: Matt Eagleson <e.Messiah@superpimp.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA
 */

#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <dirent.h>
#include "pan.h"
#include "../config.h"

/* Solaris' dirent.d_nam can be up to MAXNAMLEN */
#ifdef DIRENT_NEEDS_MAXNAMLEN
#define D_NAME_MAXLEN MAXNAMLEN
#else
#define D_NAME_MAXLEN NAME_MAX
#endif


void
acache_delete (const char *message_id)
{
	gchar* path;
			
#if SHOW_DEBUG_MSG
	g_print ("Cache delete: %s\n", message_id);
#endif
	path = g_strdup_printf ( "/%s/%s.head", data_dir, message_id);
	unlink (path);
	g_free (path);
	path = g_strdup_printf ( "/%s/%s.body", data_dir, message_id);
	unlink (path);
	g_free (path);
	path = g_strdup_printf ( "/%s/k%s.head", data_dir, message_id);
	unlink (path);
	g_free (path);
	path = g_strdup_printf ( "/%s/k%s.body", data_dir, message_id);
	unlink (path);
	g_free (path);
}


void
acache_expire (void)
{
	char *p;
	char dirpath[PATH_MAX];
	char path[PATH_MAX];
	DIR *dir_p;
	struct dirent *dirent_p;
	struct stat stat_p;
	time_t t_oldest;
	char n_oldest[D_NAME_MAXLEN] = "";
	int cache_max, cache_size, n;

	cache_max = gnome_config_get_int_with_default("/Pan/Cache/MaxSize=1242880", NULL);
	g_snprintf (dirpath, PATH_MAX, "/%s", data_dir);
	cache_size = cache_max + 1;
	
	t_oldest = n = 0;
	while (cache_max < cache_size) {
		/* If it's too big and there is more than one entry, delete one */
		if ((cache_max < cache_size) && (n > 1) && t_oldest) { 
			p = (char *)strchr(n_oldest, '>');
			p[1] = '\0';
			acache_delete(n_oldest);
		}
		
		cache_size = t_oldest = n = 0;
		
		/* Get the size and oldest file */
   		dir_p = opendir(dirpath);
		while((dirent_p = readdir(dir_p))) {
			if ((dirent_p->d_name[0] != 'k') && (strstr(dirent_p->d_name, ".head") || strstr(dirent_p->d_name, ".body"))) {
				g_snprintf(path, PATH_MAX, "/%s/%s", data_dir, dirent_p->d_name);
				if ((stat(path, &stat_p) == 0) && (S_ISREG(stat_p.st_mode))) {
					cache_size += stat_p.st_size;
					n++;
					if ((stat_p.st_mtime < t_oldest) || (t_oldest == 0)) {
						strcpy(n_oldest, dirent_p->d_name);
						t_oldest = stat_p.st_mtime;
					} 
				}
			}
		}	
		closedir (dir_p);
#if SHOW_DEBUG_MSG
		g_print ("Cache size: %d\n", cache_size);
#endif
	}
}


FILE *
acache_open_header (article_data *adata, int flag)
{
	FILE *f;
	char path[PATH_MAX];

	if (flag == ACACHE_KEEP)	
		g_snprintf (path, PATH_MAX, "/%s/k%s.head", data_dir, adata->message_id);
	else
		g_snprintf (path, PATH_MAX, "/%s/%s.head", data_dir, adata->message_id);
	f = fopen (path, "w");
	return f;
}


void
acache_putline_header (FILE *f, char *line)
{
	fputs (line, f);
}


void
acache_close_header (FILE *f)
{
	fclose (f);
}


FILE *
acache_open_body (article_data *adata, int flag)
{
	FILE *f;
	char path[PATH_MAX];
	
	if (flag == ACACHE_KEEP)	
		g_snprintf (path, PATH_MAX, "/%s/k%s.body", data_dir, adata->message_id);
	else
		g_snprintf (path, PATH_MAX, "/%s/%s.body", data_dir, adata->message_id);

	f = fopen (path, "w");
	return f;
}


void
acache_putline_body (FILE *f, char *line)
{
	fputs (line, f);
}


void
acache_close_body (FILE *f)
{
	int cache_max;

	fclose(f);
	
	cache_max = gnome_config_get_int ("/Pan/Cache/MaxSize=5242880");
	/* This will expire some, if necessary */
	acache_expire();
}

static gchar*
acache_get_existing_filename (const char *message_id, const char* suffix)
{
	struct stat st;
	gchar* filename = g_strdup_printf ( "/%s/%s.%s", data_dir, message_id, suffix );
	int rc = stat ( filename, &st );
	if ( !rc )
		return filename;

	g_free ( filename );
	filename = g_strdup_printf ( "/%s/k%s.%s", data_dir, message_id, suffix );
	rc = stat ( filename, &st );
	if ( !rc )
		return filename;

	g_free ( filename );
	return NULL;
}

gboolean
acache_body_exists (const char *message_id)
{
	gchar* filename = acache_get_existing_filename(message_id, "body");
	const gboolean filename_found = filename!=NULL;
	g_free(filename);
	return filename_found;
}

int
acache_header_exists (const char *message_id)
{
	gchar* filename = acache_get_existing_filename(message_id, "head");
	const gboolean filename_found = filename!=NULL;
	g_free(filename);
	return filename_found;
}


static char*
acache_load_something (const char *message_id, const char* suffix)
{
	char *buf = NULL;

	/* get the filename... */
	gchar* filename = acache_get_existing_filename (message_id, suffix);
	if ( filename ) {
		GArray* file = read_file ( filename );
		if ( file ) {
			buf = file->data;
			g_array_free ( file, FALSE );
		}
		g_free ( filename );
	}

	return buf;
}
char*
acache_load_body (const char *message_id)
{
	return acache_load_something(message_id, "body");
}
char*
acache_load_header (const char *message_id)
{
	return acache_load_something(message_id, "head");
}
